""" File upload/download widgets
"""
# pylint: disable=no-member
import json

from .base import HTMLBase, T, W


@W.register
class File(HTMLBase):
    """A file. Might be uploaded from the browser.

    https://developer.mozilla.org/en-US/docs/Web/HTML/Element/input/file
    """

    _model_name = T.Unicode("FileModel").tag(sync=True)
    _view_name = T.Unicode("FileView").tag(sync=True)

    value = T.Bytes(help="The binary representation of the file").tag(
        sync=True, **W.trait_types.bytes_serialization
    )
    name = T.Unicode(help="The file's name").tag(sync=True)
    last_modified = T.Int(help="Timestamp of last file modification").tag(sync=True)
    size = T.Int(help="The size of the file in bytes").tag(sync=True)
    mime_type = T.Unicode(help="The file's MIME type.").tag(sync=True)

    def proxy(self, klass, **kwargs):
        """Generate a Proxy file object, given a File-compatible class"""
        traits = ["name", "mime_type", "size", "last_modified", "value"]
        for trait in traits:
            if trait not in kwargs:
                kwargs[trait] = getattr(self, trait)

        klass_file = klass(**kwargs)

        with klass_file.hold_trait_notifications():
            with self.hold_trait_notifications():
                for trait in traits:
                    T.link((self, trait), (klass_file, trait))

        return klass_file


@W.register
class TextFile(File):
    """A Text file"""

    _model_name = T.Unicode("TextFileModel").tag(sync=True)

    text = T.Unicode().tag(sync=True)
    encoding = T.Unicode("utf-8")

    def __init__(self, *args, **kwargs):
        if "encoding" not in kwargs:
            kwargs["encoding"] = "utf-8"
        if "text" in kwargs and "value" not in kwargs:
            kwargs["value"] = kwargs["text"].decode(kwargs["encoding"])
        elif "value" in kwargs and "json" not in kwargs:
            kwargs["text"] = kwargs["value"].decode(encoding=kwargs["encoding"])
        super().__init__(*args, **kwargs)

        with self.hold_trait_notifications():
            self.observe(self._on_bytes, "value")
            self.observe(self._on_text, "text")

    def _on_bytes(self, change):
        self.text = change.new.decode(encoding=self.encoding)

    def _on_text(self, change):
        value = change.new.encode(self.encoding)
        size = len(value)
        if value is not None and value != self.value:
            with self.hold_trait_notifications():
                self.value, self.size = value, size


@W.register
class JSONFile(File):
    """A JSON file"""

    _model_name = T.Unicode("JSONFileModel").tag(sync=True)

    json = T.Union(
        [T.Dict(), T.List(), T.Unicode(), T.Int(), T.Float(), T.Bool()], allow_none=True
    ).tag(sync=True)

    def __init__(self, *args, **kwargs):
        if "json" in kwargs and "value" not in kwargs:
            try:
                kwargs["value"] = json.dumps(
                    kwargs["json"], indent=2, sort_keys=True
                ).encode("utf-8")
            except json.JSONDecodeError:
                pass
        elif "value" in kwargs and "json" not in kwargs:
            try:
                kwargs["json"] = json.loads(kwargs["value"].decode(encoding="utf-8"))
            except json.JSONDecodeError:
                pass
        super().__init__(*args, **kwargs)

        with self.hold_trait_notifications():
            self.observe(self._on_bytes, "value")
            self.observe(self._on_json, "json")

    def _on_bytes(self, change):
        self.json = json.loads(change.new.decode(encoding="utf-8"))

    def _on_json(self, change):
        value = json.dumps(change.new, sort_keys=True, indent=2).encode("utf-8")
        size = len(value)
        if value is not None and value != self.value:
            with self.hold_trait_notifications():
                self.value, self.size = value, size


@W.register
class FileBox(HTMLBase, W.Box):
    """A box of files, which can be used to upload and download files"""

    _model_name = T.Unicode("FileBoxModel").tag(sync=True)
    _view_name = T.Unicode("FileBoxView").tag(sync=True)

    children = W.trait_types.TypedTuple(
        trait=T.Instance(File), help="List of file widgets"
    ).tag(sync=True, **W.widget_serialization)

    accept = W.trait_types.TypedTuple(
        trait=T.Unicode(), help="uploadable extensions and mimetypes (with wildcards)"
    ).tag(sync=True)

    multiple = T.Bool(False, help="if true, accept multiple files").tag(sync=True)



""" Base classes for HTML 5
"""
# pylint: disable=unused-import
import ipywidgets as W  # noqa
import traitlets as T

from wxyz.core.base import Base

from ._version import module_name, module_version


class HTMLBase(Base):
    """Module metadata for HTML5"""

    _model_module = T.Unicode(module_name).tag(sync=True)
    _model_module_version = T.Unicode(module_version).tag(sync=True)
    _view_module = T.Unicode(module_name).tag(sync=True)
    _view_module_version = T.Unicode(module_version).tag(sync=True)



""" some constants
"""
# pylint: disable=invalid-name
import json
from pathlib import Path

HERE = Path(__file__).parent
__jspackage__ = json.loads((HERE / "js/package.json").read_text(encoding="utf-8"))
__version__ = __jspackage__["version"]
module_name = __jspackage__["name"]
module_version = f"""^{__jspackage__["version"]}"""



""" Full Screen widget
"""
from .base import HTMLBase, T, W


@W.register
class Fullscreen(HTMLBase, W.Box):
    """A full-screen Box, activated with Ctrl+Shift+Click, or the flower key on MacOS"""

    _model_name = T.Unicode("FullscreenModel").tag(sync=True)
    _view_name = T.Unicode("FullscreenView").tag(sync=True)



""" Widgets for core HTML 5 controls
"""
from ._version import __version__, module_name
from .widget_color import AlphaColor, AlphaColorPicker, EmptyAlphaColor
from .widget_file import File, FileBox, JSONFile, TextFile
from .widget_fullscreen import Fullscreen

__all__ = [
    "__version__",
    "AlphaColor",
    "AlphaColorPicker",
    "EmptyAlphaColor",
    "File",
    "FileBox",
    "Fullscreen",
    "JSONFile",
    "TextFile",
]


def _jupyter_labextension_paths():
    return [{"src": "labextension", "dest": module_name}]



""" Some more color traits and widgets
"""
# pylint: disable=R0903,R0901
import re

from wxyz.core.base import W

RE_RGB_EH = re.compile(r"rgba?\(\d+,\s*\d+,\s*\d+,\s*[10](\.\d+)?\)")


class AlphaColor(W.widgets.trait_types.Color):
    """A color with alpha"""

    def validate(self, obj, value):
        """expand the validation to work for rgba"""
        if RE_RGB_EH.match(value):
            return value
        return super().validate(obj, value)


class EmptyAlphaColor(W.widgets.trait_types.Color):
    """A color with alpha that might be the empty string"""

    def validate(self, obj, value):
        """expand the validation to work for rgba"""
        if value == "":
            return value
        return super().validate(obj, value)


class AlphaColorPicker(W.ColorPicker):
    """A color picker that should allow setting opacity"""

    value = AlphaColor("rgba(128, 128, 128, 0.5)").tag(sync=True)



"""generated setup for wxyz_html, do not edit by hand"""
import json
from pathlib import Path

WXYZ_NAME = "html"

HERE = Path(__file__).parent
JS_PKG = HERE / f"src/wxyz/{WXYZ_NAME}/js/package.json"

__jspackage__ = json.loads(JS_PKG.read_text(encoding="utf-8"))


HERE = Path(__file__).parent
EXT_NAME = __jspackage__["name"]

EXT_FILES = {}

SHARE = f"share/jupyter/labextensions/{EXT_NAME}"

EXT = HERE / f"src/wxyz/{WXYZ_NAME}/labextension"

for ext_path in [EXT] + [d for d in EXT.rglob("*") if d.is_dir()]:
    if ext_path == EXT:
        target = str(SHARE)
    else:
        target = f"{SHARE}/{ext_path.relative_to(EXT)}"
    EXT_FILES[target] = [
        str(p.relative_to(HERE).as_posix())
        for p in ext_path.glob("*")
        if not p.is_dir()
    ]

ALL_FILES = sum(EXT_FILES.values(), [])

assert (
    len([p for p in ALL_FILES if "remoteEntry" in str(p)]) == 1
), "expected _exactly one_ remoteEntry.*.js"

EXT_FILES[SHARE] += [f"src/wxyz/{WXYZ_NAME}/install.json"]

SETUP_ARGS = dict(
    version=__jspackage__["version"],
    data_files=[(str(k), list(map(str, v))) for k, v in EXT_FILES.items()],
)

if __name__ == "__main__":
    import setuptools

    setuptools.setup(**SETUP_ARGS)
